@@ -1,10 +1,33 @@
 #include <CoreFoundation/CoreFoundation.h>
+#include <CoreFoundation/CFBundlePriv.h>
 #include <libc.h>
 #include <Kernel/libsa/mkext.h>
 #include <sys/types.h>
 #include <sys/mman.h>
+#include <fts.h>
+#include <mach-o/arch.h>
+#include <mach-o/fat.h>
+#include <mach-o/loader.h>
+#include <mach-o/nlist.h>
+#include <mach/mach.h>
+#include <mach/mach_error.h>
+#include <mach/machine/vm_param.h>
+#include <mach/kmod.h>
+#include <mach/mach_types.h>
 
 #include <IOKit/kext/KXKextManager.h>
+#include <IOKit/IOTypes.h>
+#include <IOKit/IOKitLib.h>
+#include <IOKit/IOKitServer.h>
+#include <IOKit/IOCFUnserialize.h>
+#include <IOKit/IOCFSerialize.h>
+#include <libkern/OSByteOrder.h>
+
+
+#define DEFAULT_CACHE_DIR	"/System/Library/Caches/com.apple.kernelcaches"
+#define SA_HAS_RUN_PATH		"/var/db/.AppleSetupDone"
+#define kKXROMExtensionsFolder  "/System/Library/Caches/com.apple.romextensions/"
+#define TEMP_DIR		"/com.apple.iokit.kextcache.mkext.XX"
 
 /*******************************************************************************
 * Global variables.
@@ -17,10 +40,8 @@ char * progname = "(unknown)";
 *******************************************************************************/
 
 // In compression.c
-__private_extern__ u_int8_t *
-encodeLZSS(u_int8_t *dstP, long dstLen, u_int8_t *srcP, long srcLen);
-__private_extern__ void
-checkLZSS(u_int8_t *codeP, u_int8_t *srcEnd, u_int8_t *textP, u_int32_t tLen);
+__private_extern__ u_int32_t
+local_adler32(u_int8_t *buffer, int32_t length);
 
 // In arch.c
 __private_extern__ void
@@ -30,12 +51,28 @@ __private_extern__ int
 get_arch_from_flag(char *name, cpu_type_t *cpuP, cpu_subtype_t *subcpuP);
 
 // in mkext_file.c
-Boolean createMkextArchive(CFDictionaryRef kextDict,
+ssize_t createMkextArchive(
+    int fd,
+    CFDictionaryRef kextDict,
     const char * mkextFilename,
     const char * archName,
     cpu_type_t archCPU,
     cpu_subtype_t archSubtype,
     int verbose_level);
+Boolean checkMkextArchiveSize( ssize_t size );
+
+// in prelink.c
+KXKextManagerError
+PreLink(KXKextManagerRef theKextManager, CFDictionaryRef kextDict,
+        const char * kernelCacheFilename,
+	const struct timeval *cacheFileTimes,
+	const char * kernel_file_name, 
+	const char * platform_name,
+	const char * root_path,
+	CFSetRef kernel_requests,
+	Boolean all_plists,
+	const NXArchInfo * arch,
+	int verbose_level, Boolean debug_mode);
 
 /*******************************************************************************
 * Utility and callback functions.
@@ -48,6 +85,8 @@ static void addKextForMkextCache(
     cpu_subtype_t archSubtype,
     Boolean local,
     Boolean network,
+    Boolean safeboot,
+    CFSetRef kernel_requests,
     int verbose_level,
     Boolean do_tests);
 static void collectKextsForMkextCache(
@@ -57,10 +96,12 @@ static void collectKextsForMkextCache(
     cpu_subtype_t archSubtype,
     Boolean local,
     Boolean network,
+    Boolean safeboot,
+    CFSetRef kernel_requests,
     int verbose_level,
     Boolean do_tests);
 
-static void verbose_log(const char * format, ...);
+__private_extern__ void verbose_log(const char * format, ...);
 static void error_log(const char * format, ...);
 static int user_approve(int default_answer, const char * format, ...);
 static const char * user_input(const char * format, ...);
@@ -69,8 +110,14 @@ static Boolean addKextsToManager(
     CFArrayRef kextNames,
     CFMutableArrayRef kextNamesToUse,
     Boolean do_tests);
+static void get_catalog_demand_lists(CFMutableSetRef * kernel_requests,
+				      CFSetRef * kernel_cache_misses,
+				      int verbose_level);
 static void usage(int level);
 
+#define kMaxArchs 64
+#define kRootPathLen 256
+
 /*******************************************************************************
 *******************************************************************************/
 
@@ -90,22 +137,48 @@ int main(int argc, const char * argv[])
     Boolean forkExit = false;                // -F
     Boolean local_for_repositories = false;  // -l
     Boolean network_for_repositories = false;// -n
+    Boolean safeboot_for_repositories = false; // -s
     Boolean local_for_all = false;           // -L
     Boolean network_for_all = false;         // -N
-#if 0 // unsupported
-    Boolean patch_vtables = false;           // -p
-#endif 0
+    Boolean safeboot_for_all = false;        // -S
     Boolean repositoryCaches = false;        // -k
-    char * mkextFilename = NULL;             // -m; don't release
+    const char * mkextFilename = NULL;       // -m; don't release
+    char kernelCacheBuffer[PATH_MAX];
+    const char * kernelCacheFilename = NULL; // -c; don't release
     Boolean pretend_authentic = false;       // -z
+    Boolean load_in_task = false;
+    Boolean include_kernel_requests = false;
+    Boolean cache_looks_uptodate = false;
+    Boolean debug_mode = false;		     // -d
+    int verbose_level = 0;		     // -v
+    const char * default_kernel_file = "/mach_kernel";
+    const char * kernel_file = default_kernel_file;  // overriden by -K option
+    const char * source_extensions    = "/System/Library/Extensions";
+    int fd = -1;
+    const char * output_filename = NULL;
+    const char * temp_dir;
+    char * temp_file = NULL;
+
+    struct stat kernel_stat_buf;
+    struct stat extensions_stat_buf;
+    struct stat rom_extensions_stat_buf;
+    Boolean have_kernel_time, have_extensions_time;
+    struct timeval _cacheFileTimes[2];
+    struct timeval *cacheFileTimes = NULL;
+
+    struct {
+	char platform_name[64];
+	char root_path[kRootPathLen];
+    } platform_name_root_path;
+    io_registry_entry_t entry;
+    Boolean all_plists;
 
     // -a for these three
-    Boolean archSet = false;
-    const char * archName = "any";
-    cpu_type_t archCPU = CPU_TYPE_ANY;
-    cpu_subtype_t archSubtype = CPU_SUBTYPE_MULTIPLE;
-
-    int verbose_level = 0;                   // -v
+    NXArchInfo archAny = {
+        "any",
+        CPU_TYPE_ANY,
+        CPU_SUBTYPE_MULTIPLE
+    };
 
     CFMutableArrayRef repositoryDirectories = NULL;  // must release
     CFMutableArrayRef repositories = NULL;           // must release
@@ -114,6 +187,10 @@ int main(int argc, const char * argv[])
     CFMutableArrayRef namedKexts = NULL;             // must release
     CFMutableArrayRef repositoryKexts = NULL;        // must release
     CFMutableDictionaryRef checkDictionary = NULL;   // must release
+    CFMutableSetRef   kernel_requests = NULL;	     // must release
+    CFSetRef	      kernel_cache_misses = NULL;    // must release
+    const NXArchInfo *archs[kMaxArchs] = {NULL};
+    int nArchs = 0;
 
    /*****
     * Find out what the program was invoked as.
@@ -186,50 +263,153 @@ int main(int argc, const char * argv[])
         goto finish;
     }
 
+    bzero(&platform_name_root_path, sizeof(platform_name_root_path));
+    entry = IORegistryEntryFromPath(kIOMasterPortDefault, kIOServicePlane ":/");
+    if (entry)
+    {
+	if (KERN_SUCCESS != IORegistryEntryGetName(entry, platform_name_root_path.platform_name))
+	    platform_name_root_path.platform_name[0] = 0;
+	IOObjectRelease(entry);
+    }
 
+    entry = IORegistryEntryFromPath(kIOMasterPortDefault, kIODeviceTreePlane ":/chosen");
+    if (entry)
+    {
+	CFTypeRef obj = 0;
+	obj = IORegistryEntryCreateCFProperty(entry, CFSTR("rootpath"), kCFAllocatorDefault, kNilOptions);
+	if (obj && (CFGetTypeID(obj) == CFDataGetTypeID()))
+	{
+	    CFIndex len = CFDataGetLength((CFDataRef) obj);
+	    strncpy(platform_name_root_path.root_path, (char *)CFDataGetBytePtr((CFDataRef) obj), len);
+	    platform_name_root_path.root_path[len] = 0;
+	} else {
+            const char *data;
+            char *ptr = platform_name_root_path.root_path;
+            CFIndex len;
+            // Construct entry from UUID of boot volume and kernel name.
+            obj = 0;
+            do {
+                obj = IORegistryEntryCreateCFProperty(entry, CFSTR("boot-device-path"), kCFAllocatorDefault, kNilOptions);
+                if (!obj)
+                    break;
+
+                if (CFGetTypeID(obj) == CFDataGetTypeID()) {
+                    data = (char *)CFDataGetBytePtr((CFDataRef) obj);
+                    len = CFDataGetLength((CFDataRef) obj);
+                } else if (CFGetTypeID(obj) == CFStringGetTypeID()) {
+                    data = CFStringGetCStringPtr((CFStringRef) obj, kCFStringEncodingUTF8);
+                    if (!data)
+                        break;
+                    len = strlen(data) + 1; // include trailing null
+                } else {
+                    break;
+                }
+                if (len > kRootPathLen)
+                    len = kRootPathLen;
+                memcpy(ptr, data, len);
+                ptr += len;
+
+                CFRelease(obj);
+
+                obj = IORegistryEntryCreateCFProperty(entry, CFSTR("boot-file"), kCFAllocatorDefault, kNilOptions);
+                if (!obj)
+                    break;
+
+                if (CFGetTypeID(obj) == CFDataGetTypeID()) {
+                    data = (char *)CFDataGetBytePtr((CFDataRef) obj);
+                    len = CFDataGetLength((CFDataRef) obj);
+                } else if (CFGetTypeID(obj) == CFStringGetTypeID()) {
+                    data = CFStringGetCStringPtr((CFStringRef) obj, kCFStringEncodingUTF8);
+                    if (!data)
+                        break;
+                    len = strlen(data);
+                } else {
+                    break;
+                }
+                if ((ptr - platform_name_root_path.root_path + len) >= kRootPathLen)
+                    len = kRootPathLen - (ptr - platform_name_root_path.root_path);
+                memcpy(ptr, data, len);
+            } while (0);
+        }
+        if (obj)
+            CFRelease(obj);
+	IOObjectRelease(entry);
+    }
+    if (!platform_name_root_path.platform_name[0] || !platform_name_root_path.root_path[0])
+	platform_name_root_path.platform_name[0] = platform_name_root_path.root_path[0] = 0;
+
+    have_kernel_time     = (0 == stat(kernel_file,       &kernel_stat_buf));
+    have_extensions_time = (0 == stat(source_extensions, &extensions_stat_buf));
+
+    if (have_kernel_time || have_extensions_time)
+    {
+	cacheFileTimes = _cacheFileTimes;
+	if (!have_kernel_time 
+	    || (have_extensions_time && (extensions_stat_buf.st_mtime > kernel_stat_buf.st_mtime)))
+	{
+	    TIMESPEC_TO_TIMEVAL(&cacheFileTimes[0], &extensions_stat_buf.st_atimespec);
+	    TIMESPEC_TO_TIMEVAL(&cacheFileTimes[1], &extensions_stat_buf.st_mtimespec);
+	}
+	else
+	{
+	    TIMESPEC_TO_TIMEVAL(&cacheFileTimes[0], &kernel_stat_buf.st_atimespec);
+	    TIMESPEC_TO_TIMEVAL(&cacheFileTimes[1], &kernel_stat_buf.st_mtimespec);
+	}
+	cacheFileTimes[1].tv_sec++;
+    }
     /*****
     * Process command line arguments. If running in kextload-compatibiliy
     * mode, accept its old set of options and none other. If running in
     * the new mode, process the new, larger set of options.
     */
     while ((optchar = getopt(argc, (char * const *)argv,
-               "a:eFklLm:nNptvz")) != -1) {
+               "a:cdeFhkK:lLm:nNrsStvz")) != -1) {
 
         switch (optchar) {
 
           case 'a':
-            if (archSet) {
-                fprintf(stderr, "duplicate use of -%c option\n\n", optchar);
-                usage(0);
-                exit_code = 1;
-                goto finish;
+            if (nArchs >= kMaxArchs) {
+                  fprintf(stderr, "maximum of %d architectures supported\n", kMaxArchs);
+                  exit_code = 1;
+                  goto finish;
             }
-            if (!get_arch_from_flag(optarg, &archCPU, &archSubtype)) {
+            archs[nArchs] = NXGetArchInfoFromName(optarg);
+            if (archs[nArchs] == NULL) {
                 fprintf(stderr, "unknown CPU arch %s\n\n", optarg);
                 usage(0);
                 exit_code = 1;
                 goto finish;
             }
-            archName = optarg;
-            archSet = true;
+            ++nArchs;
             break;
 
+          case 'r':
+	    include_kernel_requests = true;
+	    break;
+
+          case 'd':
+	    debug_mode = true;
+	    break;
+
           case 'e':
             if (mkextFilename) {
-                fprintf(stderr, "an mkext filename has already been set\n\n");
+                fprintf(stderr, "an output filename has already been set\n\n");
                 usage(0);
                 exit_code = 1;
                 goto finish;
             }
             mkextFilename = "/System/Library/Extensions.mkext";
-            CFArrayAppendValue(repositoryDirectories,
-                CFSTR("/System/Library/Extensions"));
+            CFArrayAppendValue(repositoryDirectories, kKXSystemExtensionsFolder);
             break;
 
           case 'F':
             forkExit = true;
             break;
 
+          case 'K':
+            kernel_file = optarg;
+            break;
+
           case 'k':
             repositoryCaches = true;
             break;
@@ -242,6 +422,72 @@ int main(int argc, const char * argv[])
             local_for_all = true;
             break;
 
+          case 'c':
+            if (kernelCacheFilename) {
+                fprintf(stderr, "an output filename has already been set\n\n");
+                usage(0);
+                exit_code = 1;
+                goto finish;
+            }
+
+	    if (optind >= argc) {
+
+		// default args for kextd's usage
+
+		kernelCacheFilename = DEFAULT_CACHE_DIR "/kernelcache";
+		struct stat cache_stat_buf;
+
+		if ((-1 == stat(DEFAULT_CACHE_DIR, &cache_stat_buf))
+		    || !(S_IFDIR & cache_stat_buf.st_mode)) {
+		    mkdir(DEFAULT_CACHE_DIR, 0755);
+		}
+		// Make sure we scan the standard Extensions folder.
+		CFArrayAppendValue(repositoryDirectories,
+		    kKXSystemExtensionsFolder);
+		// Make sure we scan the ROM Extensions folder.
+		if (0 == stat(kKXROMExtensionsFolder, &rom_extensions_stat_buf))
+		    CFArrayAppendValue(repositoryDirectories, CFSTR(kKXROMExtensionsFolder));
+
+		if (!platform_name_root_path.platform_name[0]) {
+		    // need more than the minimal set
+		    local_for_repositories = true;
+		}
+
+		if (!include_kernel_requests) {
+		    // this cache isn't tied to a config
+		    platform_name_root_path.root_path[0] = platform_name_root_path.platform_name[0] = 0;
+		}
+	
+		if (platform_name_root_path.platform_name[0] || platform_name_root_path.root_path[0])
+		{
+		    sprintf(kernelCacheBuffer, "%s.%08X", 
+			kernelCacheFilename,
+			NXSwapHostIntToBig(local_adler32((u_int8_t *) &platform_name_root_path, 
+					    sizeof(platform_name_root_path))));
+		    kernelCacheFilename = kernelCacheBuffer;
+		}
+
+		if ((0 == stat(kernelCacheFilename, &cache_stat_buf))
+		    && have_kernel_time && have_extensions_time)
+		{
+		    if ((cache_stat_buf.st_mtime > kernel_stat_buf.st_mtime)
+  		     &&  (cache_stat_buf.st_mtime > extensions_stat_buf.st_mtime)
+		     &&  (cache_stat_buf.st_mtime == cacheFileTimes[1].tv_sec))
+			cache_looks_uptodate = true;
+		}
+
+		if (-1 == stat(SA_HAS_RUN_PATH, &cache_stat_buf)) {
+		    if (verbose_level >= 1) {
+			verbose_log("SetupAssistant not yet run");
+		    }
+		    exit(0);
+		}
+
+	    } else {
+		kernelCacheFilename = argv[optind++];
+	    }
+            break;
+
           case 'm':
             if (mkextFilename) {
                 fprintf(stderr, "an mkext filename has already been set\n\n");
@@ -260,11 +506,14 @@ int main(int argc, const char * argv[])
             network_for_all = true;
             break;
 
-#if 0 //unsupported
-          case 'p':
-            patch_vtables = true;
+
+          case 's':
+            safeboot_for_repositories = true;
+            break;
+
+          case 'S':
+            safeboot_for_all = true;
             break;
-#endif 0
 
           case 't':
             do_tests = true;
@@ -302,6 +551,11 @@ int main(int argc, const char * argv[])
             pretend_authentic = true;
             break;
 
+	case 'h':
+	  usage(2);
+	  exit_code = 0;
+	  goto finish;
+
         default:
             fprintf(stderr, "unknown option -%c\n", optchar);
             usage(0);
@@ -318,8 +572,8 @@ int main(int argc, const char * argv[])
    /*****
     * Check for bad combinations of options.
     */
-    if (!mkextFilename && !repositoryCaches) {
-        fprintf(stderr, "no work to do; one of -m or -k must be specified\n");
+    if (!mkextFilename && !kernelCacheFilename && !repositoryCaches) {
+        fprintf(stderr, "no work to do; one of -m, -c or -k must be specified\n");
         usage(0);
         exit_code = 1;
         goto finish;
@@ -341,7 +595,30 @@ int main(int argc, const char * argv[])
         if (CFStringHasSuffix(argString, CFSTR(".kext")) ||
             CFStringHasSuffix(argString, CFSTR(".kext/")) ) {
             CFArrayAppendValue(kextNames, argString);
-        } else {
+        }
+        else
+        {
+            if (!CFArrayGetCount(repositoryDirectories))
+            {
+                source_extensions = argv[i];
+                have_extensions_time = (0 == stat(source_extensions, &extensions_stat_buf));
+                if (have_kernel_time || have_extensions_time)
+                {
+                    cacheFileTimes = _cacheFileTimes;
+                    if (!have_kernel_time 
+                        || (have_extensions_time && (extensions_stat_buf.st_mtime > kernel_stat_buf.st_mtime)))
+                    {
+                        TIMESPEC_TO_TIMEVAL(&cacheFileTimes[0], &extensions_stat_buf.st_atimespec);
+                        TIMESPEC_TO_TIMEVAL(&cacheFileTimes[1], &extensions_stat_buf.st_mtimespec);
+                    }
+                    else
+                    {
+                        TIMESPEC_TO_TIMEVAL(&cacheFileTimes[0], &kernel_stat_buf.st_atimespec);
+                        TIMESPEC_TO_TIMEVAL(&cacheFileTimes[1], &kernel_stat_buf.st_mtimespec);
+                    }
+                    cacheFileTimes[1].tv_sec++;
+                }
+            }
             CFArrayAppendValue(repositoryDirectories, argString);
         }
         CFRelease(argString);
@@ -369,6 +646,19 @@ int main(int argc, const char * argv[])
             break;
           case 0:   // child task
             setpriority(PRIO_PROCESS, getpid(), 20); // run at really low priority
+	    if (include_kernel_requests) {
+		kern_return_t    kern_result;
+		mach_timespec_t  waitTime = { 40, 0 };
+
+		kern_result = IOKitWaitQuiet(kIOMasterPortDefault, &waitTime);
+		if (kern_result == kIOReturnTimeout) {
+		    fprintf(stderr, "IOKitWaitQuiet() timed out");
+		} else if (kern_result != kIOReturnSuccess) {
+		    fprintf(stderr, "IOKitWaitQuiet() failed with result code %x",
+				    kern_result);
+		}
+		sleep(40);
+	    }
             break;
           default:  // parent task
             exit_code = 0;
@@ -377,9 +667,28 @@ int main(int argc, const char * argv[])
         }
     }
 
+    if (include_kernel_requests)
+	get_catalog_demand_lists(&kernel_requests, &kernel_cache_misses, verbose_level);
+
+    if (cache_looks_uptodate)
+    {
+	CFShow(kernel_cache_misses);
+	CFShow(kernel_requests);
+	if ((!kernel_cache_misses || !kernel_requests)
+	    || (!CFSetGetCount(kernel_cache_misses))
+	    || (CFSetGetCount(kernel_cache_misses) == CFSetGetCount(kernel_requests)))
+	{
+	    if (verbose_level >= 1) {
+		verbose_log("cache %s up to date", kernelCacheFilename);
+	    }
+	    exit(0);
+	}
+    }
+
    /*****
     * Set up the kext manager.
     */
+
     theKextManager = KXKextManagerCreate(kCFAllocatorDefault);
     if (!theKextManager) {
         fprintf(stderr, "can't allocate kernel extension manager\n");
@@ -387,7 +696,7 @@ int main(int argc, const char * argv[])
         goto finish;
     }
 
-    result = KXKextManagerInit(theKextManager, false /* load_in_task */,
+    result = KXKextManagerInit(theKextManager, load_in_task,
         false /* safeBoot */);
     if (result != kKXKextManagerErrorNone) {
         fprintf(stderr, "can't initialize kernel extension manager (%s)\n",
@@ -397,6 +706,7 @@ int main(int argc, const char * argv[])
     }
 
     KXKextManagerSetPerformsFullTests(theKextManager, do_tests);
+    KXKextManagerSetPerformsStrictAuthentication(theKextManager, true);
     KXKextManagerSetLogLevel(theKextManager, verbose_level);
     KXKextManagerSetLogFunction(theKextManager, &verbose_log);
     KXKextManagerSetErrorLogFunction(theKextManager, &error_log);
@@ -427,7 +737,15 @@ int main(int argc, const char * argv[])
         directoryURL = CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
                 directory, kCFURLPOSIXPathStyle, true);
         if (!directoryURL) {
-            fprintf(stderr, "memory allocation failure\n");
+            char repositoryPath[MAXPATHLEN];
+            if (!CFStringGetCString(directory, repositoryPath,
+                sizeof(repositoryPath) / sizeof(char), kCFStringEncodingMacRoman)) {
+
+                fprintf(stderr, "string conversion error\n");
+
+            } else {
+                fprintf(stderr, "'%s': no such directory\n", repositoryPath);
+            }
             exit_code = 1;
 
             goto finish;
@@ -439,8 +757,7 @@ int main(int argc, const char * argv[])
         if (result != kKXKextManagerErrorNone) {
             fprintf(stderr, "can't add repository (%s).\n",
                 KXKextManagerErrorStaticCStringForError(result));
-            exit_code = 1;
-            goto finish;
+	    continue;
         }
 
         CFArrayAppendValue(repositories, repository);
@@ -494,21 +811,19 @@ int main(int argc, const char * argv[])
         for (i = 0; i < count; i++) {
             KXKextRepositoryRef repository = (KXKextRepositoryRef)
                 CFArrayGetValueAtIndex(repositories, i);
-            KXKextManagerError kmResult = KXKextRepositoryWriteCache(
+            /*KXKextManagerError kmResult = */ KXKextRepositoryWriteCache(
                 repository, NULL);
             // FIXME: Do anything with kmResult?
         }
     }
 
    /*****
-    * Do the mkext cache file.
+    * Do the cache files.
     */
-    if (!mkextFilename) {
+    if (!mkextFilename && !kernelCacheFilename) {
         goto finish;
     }
 
-    // FIXME: Check that mkextFilename can be created & written to.
-
    /*****
     * Get KXKextRef objects for each of the kexts named on the command line.
     */
@@ -538,30 +853,227 @@ int main(int argc, const char * argv[])
         CFArrayAppendValue(namedKexts, thisKext);
     }
 
-    collectKextsForMkextCache(namedKexts, checkDictionary,
-        archCPU, archSubtype,
-        local_for_all, network_for_all,  
-        verbose_level, do_tests);
-    collectKextsForMkextCache(repositoryKexts, checkDictionary,
-        archCPU, archSubtype, local_for_repositories || local_for_all,
-        network_for_repositories || network_for_all,
-        verbose_level, do_tests);
+    /* If no architectures are specified on the command line,
+     * create one kext file that contains everything.
+     * If one architecture is specified, create one mkext file
+     * that is thinned to the specified architecture.
+     * If multiple architectures are specified,
+     * create a fat file containing multiple thin mkext files.
+     */
 
-    if (!createMkextArchive(checkDictionary, mkextFilename,
-        archName, archCPU, archSubtype, verbose_level)) {
-        exit_code = 1;
+    if (nArchs == 0) {
+        archs[nArchs++] = &archAny;
+    }
+
+    if (kernelCacheFilename)
+    {
+	KXKextManagerError err;
+	const NXArchInfo * hostArch;
+	
+	hostArch = NXGetLocalArchInfo();
+	if (!hostArch) {
+	    fprintf(stderr, "can't determine the host architecture\n");
+	    exit_code = 1;
+	    // still create mkext if requested
+	    goto domkext;
+	}
+
+	if((nArchs > 1) ||
+	  ((archs[0]->cputype != CPU_TYPE_ANY) && (archs[0]->cputype != hostArch->cputype))) {
+	    fprintf(stderr, "can't create fat cache\n");
+	    exit_code = 1;
+	    // still create mkext if requested
+	    goto domkext;
+	}
+
+        CFDictionaryRemoveAllValues(checkDictionary);
+        collectKextsForMkextCache(namedKexts, checkDictionary,
+                                  hostArch->cputype, hostArch->cpusubtype,
+                                  local_for_all, network_for_all, safeboot_for_all,
+				  kernel_requests,
+                                  verbose_level, do_tests);
+        collectKextsForMkextCache(repositoryKexts, checkDictionary,
+                                  hostArch->cputype, hostArch->cpusubtype,
+                                  local_for_repositories || local_for_all,
+                                  network_for_repositories || network_for_all,
+                                  safeboot_for_repositories || safeboot_for_all,
+				  kernel_requests,
+                                  verbose_level, do_tests);
+
+	all_plists = false;
+	if (!include_kernel_requests) {
+	    // this cache isn't tied to a config
+	    platform_name_root_path.root_path[0] = platform_name_root_path.platform_name[0] = 0;
+	    all_plists = !(    local_for_repositories || local_for_all
+			    || network_for_repositories || network_for_all
+			    || safeboot_for_repositories || safeboot_for_all);
+	}
+	err = PreLink(theKextManager, checkDictionary, kernelCacheFilename, cacheFileTimes, kernel_file,
+			platform_name_root_path.platform_name, platform_name_root_path.root_path,
+			kernel_requests, all_plists,
+			hostArch, // archs[0],
+			verbose_level, debug_mode);
+
+	if (kKXKextManagerErrorNone != err)
+	{
+	    exit_code = 1;
+	    // still create mkext if requested
+	    goto domkext;
+	}
+    }
+
+domkext:
+    if (!mkextFilename) {
         goto finish;
     }
 
+    temp_dir = getenv("TMPDIR");
+    if (!temp_dir)
+        temp_dir = "/tmp";
+    temp_file = malloc(strlen(temp_dir) + strlen(TEMP_DIR) + 1);
+    strcpy(temp_file, temp_dir);
+    strcat(temp_file, TEMP_DIR);
+    mktemp(temp_file);
+
+    fd = open(temp_file, O_WRONLY|O_CREAT|O_TRUNC, 0666);
+    if (-1 == fd) {
+        fprintf(stderr, "can't create %s - %s\n", temp_file,
+            strerror(errno));
+        result = false;
+        goto finish;  // FIXME: mkextcache used to exit with EX_CANTCREAT
+    }
+    output_filename = temp_file;
+
+    struct fat_header fatHeader;
+    struct fat_arch fatArchs[kMaxArchs];
+    unsigned long fat_offset = 0;
+    ssize_t bytes_written = 0;
+
+    if (nArchs > 1) {
+        fat_offset = sizeof(struct fat_header) + (sizeof(struct fat_arch) * nArchs);
+        lseek(fd, fat_offset, SEEK_SET);
+    }
+
+    for (i=0; i<nArchs; i++) {
+        fatArchs[i].cputype = NXSwapHostLongToBig(archs[i]->cputype);
+        fatArchs[i].cpusubtype = NXSwapHostLongToBig(archs[i]->cpusubtype);
+        fatArchs[i].offset = NXSwapHostLongToBig(fat_offset);
+
+        if (verbose_level >= 1) {
+            verbose_log("processing architecture %s", archs[i]->name);
+        }
+        CFDictionaryRemoveAllValues(checkDictionary);
+        collectKextsForMkextCache(namedKexts, checkDictionary,
+                                  archs[i]->cputype, archs[i]->cpusubtype,
+                                  local_for_all, network_for_all, safeboot_for_all,
+				  NULL /*kernel_requests*/,
+                                  verbose_level, do_tests);
+        collectKextsForMkextCache(repositoryKexts, checkDictionary,
+                                  archs[i]->cputype, archs[i]->cpusubtype,
+                                  local_for_repositories || local_for_all,
+                                  network_for_repositories || network_for_all,
+                                  safeboot_for_repositories || safeboot_for_all,
+				  NULL /*kernel_requests*/,
+                                  verbose_level, do_tests);
+
+        bytes_written = createMkextArchive(fd, checkDictionary, output_filename,
+                                archs[i]->name, archs[i]->cputype, archs[i]->cpusubtype, verbose_level);
+
+        if (bytes_written < 0) {
+            exit_code = 1;
+            goto finish;
+        }
+        fat_offset += bytes_written;
+
+        if (checkMkextArchiveSize(bytes_written) == false) {
+                fprintf(stderr, "archive would be too large; aborting\n");
+                exit_code = 1;
+                goto finish;
+        }
+
+        fatArchs[i].size = NXSwapHostLongToBig(bytes_written);
+        fatArchs[i].align = NXSwapHostLongToBig(0);
+    }
+
+    if (nArchs > 1) {
+        lseek(fd, 0, SEEK_SET);
+        fatHeader.magic = NXSwapHostLongToBig(FAT_MAGIC);
+        fatHeader.nfat_arch = NXSwapHostLongToBig(nArchs);
+        bytes_written = write(fd, &fatHeader, sizeof(fatHeader));
+        if (bytes_written != sizeof(fatHeader)) {
+            perror("write");
+            exit_code = 1;
+            goto finish;
+        }
+
+        for (i=0; i<nArchs; i++) {
+            bytes_written = write(fd, &fatArchs[i], sizeof(fatArchs[i]));
+            if (bytes_written != sizeof(fatArchs[i])) {
+                perror("write");
+                exit_code = 1;
+                goto finish;
+            }
+        }
+    }
+
+    close(fd);
+    fd = -1;
+
+    if (have_extensions_time)
+    {
+	struct timespec mod_time = extensions_stat_buf.st_mtimespec;
+	if ((0 == stat(source_extensions, &extensions_stat_buf))
+	  && ((mod_time.tv_sec != extensions_stat_buf.st_mtimespec.tv_sec)
+	      || (mod_time.tv_nsec != extensions_stat_buf.st_mtimespec.tv_nsec)))
+	{
+	    fprintf(stderr, "cache stale - not creating %s\n", mkextFilename);
+	    exit_code = 1;
+	    goto finish;
+	}
+
+	TIMESPEC_TO_TIMEVAL(&cacheFileTimes[0], &extensions_stat_buf.st_atimespec);
+	TIMESPEC_TO_TIMEVAL(&cacheFileTimes[1], &extensions_stat_buf.st_mtimespec);
+	cacheFileTimes[1].tv_sec++;
+    }
+
+    // move it to the final destination
+    if (-1 == rename(output_filename, mkextFilename)) {
+	fprintf(stderr, "can't create file %s: %s\n", mkextFilename,
+		strerror(errno));
+	exit_code = 1;
+	goto finish;
+    }
+    output_filename = NULL;
+
+    // give the cache file the mod time of the source files when kextcache started
+    if (cacheFileTimes && (-1 == utimes(mkextFilename, cacheFileTimes))) {
+	fprintf(stderr, "can't set file times %s: %s\n", mkextFilename,
+		strerror(errno));
+	exit_code = 1;
+	goto finish;
+    }
+
 finish:
 
    /*****
     * Clean everything up.
     */
+
+    if (-1 != fd) close(fd);
+    if (output_filename)
+    {
+	if (-1 == unlink(output_filename)) {
+	    fprintf(stderr, "can't remove file %s - %s\n", output_filename,
+			strerror(errno));
+	}
+    }
+    if (temp_file)             free(temp_file);
     if (repositoryDirectories) CFRelease(repositoryDirectories);
     if (kextNames)             CFRelease(kextNames);
     if (kextNamesToUse)        CFRelease(kextNamesToUse);
     if (namedKexts)            CFRelease(namedKexts);
+    if (repositoryKexts)       CFRelease(repositoryKexts);
+    if (checkDictionary)       CFRelease(checkDictionary);
 
     if (theKextManager)        CFRelease(theKextManager);
 
@@ -569,6 +1081,42 @@ finish:
     return exit_code;
 }
 
+/*******************************************************************************
+*
+*******************************************************************************/
+
+static void get_catalog_demand_lists(CFMutableSetRef * kernel_requests,
+				      CFSetRef * kernel_cache_misses,
+				      int verbose_level)
+{
+    kern_return_t kr;
+    char * propertiesBuffer;
+    int    loaded_bytecount;
+
+    kr = IOCatalogueGetData(MACH_PORT_NULL, kIOCatalogGetModuleDemandList,
+			    &propertiesBuffer, &loaded_bytecount);
+    if (kIOReturnSuccess == kr)
+    { 
+	CFSetRef set;
+	set = (CFSetRef)
+		IOCFUnserialize(propertiesBuffer, kCFAllocatorDefault, 0, 0);
+	vm_deallocate(mach_task_self(), (vm_address_t) propertiesBuffer, loaded_bytecount);
+	*kernel_requests = CFSetCreateMutableCopy(kCFAllocatorDefault, 0, set);
+	CFRelease(set);
+    }
+
+    kr = IOCatalogueGetData(MACH_PORT_NULL, kIOCatalogGetCacheMissList,
+			    &propertiesBuffer, &loaded_bytecount);
+    if (kIOReturnSuccess == kr)
+    { 
+	if (verbose_level > 0)
+	    verbose_log("cache misses:\n%s", propertiesBuffer);
+	*kernel_cache_misses = (CFSetRef)
+		IOCFUnserialize(propertiesBuffer, kCFAllocatorDefault, 0, 0);
+	vm_deallocate(mach_task_self(), (vm_address_t) propertiesBuffer, loaded_bytecount);
+    }
+}
+
 /*******************************************************************************
 *
 *******************************************************************************/
@@ -625,6 +1173,8 @@ static void addKextForMkextCache(
     cpu_subtype_t archSubtype,
     Boolean local,
     Boolean network,
+    Boolean safeboot,
+    CFSetRef kernel_requests,
     int verbose_level,
     Boolean do_tests)
 {
@@ -679,7 +1229,21 @@ static void addKextForMkextCache(
 
     kextBundle = KXKextGetBundle(theKext);
 
-    if (local || network) {
+    include_it = false;
+    do_children = true;
+
+    if (kernel_requests) {
+	CFStringRef bundleIdentifier = CFBundleGetIdentifier(kextBundle);
+	if (!bundleIdentifier) {
+	    fprintf(stderr, "can't get identifier for kext\n");
+	    do_children = false;
+	    goto finish;
+	}
+	include_it = CFSetContainsValue(kernel_requests, bundleIdentifier);
+    } else if (!local && !network && !safeboot)
+	include_it = true;
+
+    if (!include_it && (local || network || safeboot)) {
         requiredString = CFBundleGetValueForInfoDictionaryKey(kextBundle,
             CFSTR("OSBundleRequired"));
     
@@ -716,6 +1280,13 @@ static void addKextForMkextCache(
             include_it = true;
             do_children = true;
 
+        } else if (safeboot &&
+            CFStringCompare(requiredString, CFSTR("Safe Boot"), 0) ==
+                kCFCompareEqualTo) {
+
+            include_it = true;
+            do_children = true;
+
         } else {
             char required_string[120];
 
@@ -740,10 +1311,12 @@ static void addKextForMkextCache(
         }
     }
 
-    KXKextAuthenticate(theKext);
+    if (!KXKextHasBeenAuthenticated(theKext)) {
+        KXKextAuthenticate(theKext);
+    }
 
     if (!KXKextIsAuthentic(theKext)) {
-        fprintf(stderr, "kernel extension %s is not authentic; "
+        fprintf(stderr, "kernel extension %s is not authentic (check ownership and permissions); "
              "skipping it and any plugins\n", kext_name);
         if (do_tests) {
             fprintf(stderr, "kext diagnostics:\n");
@@ -754,6 +1327,7 @@ static void addKextForMkextCache(
         goto finish;
     }
 
+    KXKextResolveDependencies(theKext);
     if (!KXKextGetHasAllDependencies(theKext)) {
         fprintf(stderr, "warning: kernel extension %s is missing dependencies "
             "(including in cache anyway; "
@@ -764,6 +1338,8 @@ static void addKextForMkextCache(
             KXKextPrintDiagnostics(theKext, stderr);
         }
         // include it anyway
+	include_it = true;
+	do_children = true;
     }
 
     if (KXKextGetDeclaresExecutable(theKext)) {
@@ -777,6 +1353,7 @@ static void addKextForMkextCache(
             do_children = true;
             goto finish;
         } else {
+
             executable_path = CFURLCopyCString(executableURL);
             if (!executable_path) {
                 fprintf(stderr, "memory allocation failure\n");
@@ -818,7 +1395,7 @@ static void addKextForMkextCache(
                 }
   
                 find_arch(NULL, &archSize, archCPU, archSubtype,
-                   machO, machOSize);
+                   (u_int8_t *)machO, machOSize);
                 if (!archSize) { // Couldn't find an architecture
                     fprintf(stderr, "%s doesn't contain code for the "
                         "architecture specified; "
@@ -848,7 +1425,7 @@ finish:
 
     if (do_children) {
         collectKextsForMkextCache(KXKextGetPlugins(theKext),
-            checkDictionary, archCPU, archSubtype, local, network,
+            checkDictionary, archCPU, archSubtype, local, network, safeboot, kernel_requests,
             verbose_level, do_tests);
     }
 
@@ -869,6 +1446,8 @@ static void collectKextsForMkextCache(
     cpu_subtype_t archSubtype,
     Boolean local,
     Boolean network,
+    Boolean safeboot,
+    CFSetRef kernel_requests,
     int verbose_level,
     Boolean do_tests)
 {
@@ -886,7 +1465,7 @@ static void collectKextsForMkextCache(
         theKext = (KXKextRef)CFArrayGetValueAtIndex(kexts, i);
 
         addKextForMkextCache(theKext, checkDictionary, archCPU, archSubtype,
-            local, network, verbose_level, do_tests);
+            local, network, safeboot, kernel_requests, verbose_level, do_tests);
     }
 
 finish:
@@ -899,7 +1478,8 @@ finish:
 *
 * Print a log message prefixed with the name of the program.
 *******************************************************************************/
-static void verbose_log(const char * format, ...)
+
+__private_extern__ void verbose_log(const char * format, ...)
 {
     va_list ap;
     char fake_buffer[2];
@@ -1183,8 +1763,9 @@ static void usage(int level)
 {
 
     fprintf(stderr,
-      "usage: %s [-a arch] [-e] [-F] [-k] [-l | -L] [-m mkext_filename]\n"
-      "    [-n | -N] [-t] [-v [1-6]] [-z] [kext_or_directory] ...\n\n",
+      "usage: %s [-a arch] [-c kernel_cache_filename] [-e] [-F] [-h] [-k]\n"
+      "    [-K kernel_filename] [-l | -L] [-r] [-m mkext_filename] [-n | -N]\n"
+      "    [-r] [-s | -S] [-t] [-v [1-6]] [-z] [kext_or_directory] ...\n\n",
       progname);
 
     if (level < 1) {
@@ -1202,12 +1783,18 @@ static void usage(int level)
         "    (required unless using -e)\n");
     fprintf(stderr,
         "  -a arch: Add only kexts that contain code for arch\n");
+    fprintf(stderr,
+        "  -c kernel_cache_filename: Create an kernel prelink cache\n");
     fprintf(stderr,
         "  -e: Create or update the mkext cache for /System/Library/Extensions\n");
     fprintf(stderr,
         "  -F: Fork and exit immediately (to work in background)\n");
+    fprintf(stderr,
+        "  -h: This usage statement\n");
     fprintf(stderr,
         "  -k: Create or update any repository .kextcache files\n");
+    fprintf(stderr,
+        "  -K kernel_filename: Name of kernel file for kernel cache\n");
     fprintf(stderr,
         "  -l: Add local-root kexts in directories to an mkext cache\n");
     fprintf(stderr,
@@ -1218,6 +1805,12 @@ static void usage(int level)
         "  -n: Add network-root kexts in directories to an mkext cache\n");
     fprintf(stderr,
         "  -N: Add network-root kexts for all args to an mkext cache\n");
+    fprintf(stderr,
+        "  -r: Add kexts previously loaded by this machine\n");
+    fprintf(stderr,
+        "  -s: Add safe boot kexts in directories to an mkext cache\n");
+    fprintf(stderr,
+        "  -S: Add safe boot kexts for all args to an mkext cache\n");
     fprintf(stderr,
         "  -t: Perform diagnostic tests on kexts and print results\n");
     fprintf(stderr,
@@ -1225,10 +1818,13 @@ static void usage(int level)
     fprintf(stderr,
         "  -z: don't authenticate kexts (for use during development)\n");
     fprintf(stderr, "\n");
-    fprintf(stderr, "At least one of -k or -m must be specified.\n");
+    fprintf(stderr, "At least one of -k, -c or -m must be specified.\n");
     fprintf(stderr,
         "-l/-L and -n/-N may both be specified to make a cache of local-\n"
         "and network-root kexts\n");
 
     return;
 }
+
+
+
