@@ -3,11 +3,14 @@
 #include <IOKit/IOKitLib.h>
 #include <IOKit/IOKitServer.h>
 #include <IOKit/IOCFURLAccess.h>
+#include <IOKit/IOCFUnserialize.h>
 #include <mach/mach.h>
 #include <mach/mach_host.h>
 #include <mach/mach_error.h>
+#include <mach-o/arch.h>
 #include <libc.h>
 #include <servers/bootstrap.h>
+#include <signal.h>
 #include <sysexits.h>
 
 #include <IOKit/kext/KXKextManager.h>
@@ -22,7 +25,10 @@
 * Globals set from invocation arguments.
 *******************************************************************************/
 
-static const char * KEXTD_SERVER_NAME = "Kernel Extension Server";
+static const char * KEXTD_SERVER_NAME = "com.apple.KernelExtensionServer";
+
+#define kKXROMExtensionsFolder        "/System/Library/Caches/com.apple.romextensions/"
+#define kKXCSystemExtensionsFolder    "/System/Library/Extensions"
 
 char * progname = "(unknown)";  // don't free
 Boolean use_repository_caches = true;
@@ -32,6 +38,10 @@ Boolean jettison_kernel_linker = true;
 int g_verbose_level = 0;        // nonzero for -v option
 Boolean safe_boot_mode = false;
 
+static Boolean parent_received_sigterm = false;
+static Boolean parent_received_sigchld = false;
+static Boolean parent_received_sigalrm = false;
+
 // options for these are not yet implemented
 char * g_kernel_file = NULL;  // don't free
 char * g_patch_dir = NULL;    // don't free
@@ -49,8 +59,11 @@ CFRunLoopRef gMainRunLoop = NULL;                      // must release
 CFRunLoopSourceRef gRescanRunLoopSource = NULL;        // must release
 CFRunLoopSourceRef gKernelRequestRunLoopSource = NULL; // must release
 CFRunLoopSourceRef gClientRequestRunLoopSource = NULL; // must release
+static CFMachPortRef gKextdSignalMachPort = NULL;      // must release
+static CFRunLoopSourceRef gSignalRunLoopSource = NULL; // must release
+
 #ifndef NO_CFUserNotification
-CFRunLoopSourceRef gNonsecureKextRunLoopSource = NULL;     // must release
+CFRunLoopSourceRef gNotificationQueueRunLoopSource = NULL;     // must release
 #endif /* NO_CFUserNotification */
 
 const char * default_kernel_file = "/mach";
@@ -59,6 +72,8 @@ queue_head_t g_request_queue;
 PTLockRef gKernelRequestQueueLock = NULL;
 PTLockRef gRunLoopSourceLock = NULL;
 
+static mach_port_t gBootstrap_port = 0;
+
 /*******************************************************************************
 * Function prototypes.
 *******************************************************************************/
@@ -70,20 +85,23 @@ static Boolean kextd_set_up_server(void);
 static void kextd_release_parent_task(void);
 
 void kextd_register_signals(void);
-void kextd_handle_sigterm(int signum);
+void kextd_parent_handle_signal(int signum);
 void kextd_handle_sighup(int signum);
-void kextd_handle_sighup_in_runloop(void * info);
+void kextd_handle_rescan(void * info);
 
+static Boolean kextd_find_rom_mkexts(void);
+static void check_extensions_mkext(void);
 static Boolean kextd_download_personalities(void);
 
-static Boolean kextd_process_ndrvs(CFArrayRef repositoryDirectories);
-
 static void usage(int level);
 
 char * CFURLCopyCString(CFURLRef anURL);
 
-__private_extern__ 
-void IOLoadPEFsFromURL( CFURLRef url, io_service_t service );
+void kextd_runloop_handle_sighup(
+    CFMachPortRef port,
+    void *msg,
+    CFIndex size,
+    void *info);
 
 /*******************************************************************************
 *******************************************************************************/
@@ -92,7 +110,8 @@ int main (int argc, const char * argv[]) {
     int exit_status = 0;
     KXKextManagerError result = kKXKextManagerErrorNone;
     int optchar;
-    CFIndex count, i;
+    CFIndex count, i, rom_repository_idx = -1;
+    Boolean have_rom_mkexts = FALSE;
 
     CFMutableArrayRef repositoryDirectories = NULL;  // -f; must free
 
@@ -126,25 +145,18 @@ int main (int argc, const char * argv[]) {
 
 #ifndef NO_CFUserNotification
 
-    gPendedNonsecureKexts = CFArrayCreateMutable(kCFAllocatorDefault, 0,
+    gPendedNonsecureKextPaths = CFArrayCreateMutable(kCFAllocatorDefault, 0,
         &kCFTypeArrayCallBacks);
-    if (!gPendedNonsecureKexts) {
+    if (!gPendedNonsecureKextPaths) {
         fprintf(stderr, "%s: memory allocation failure\n", progname);
         exit_status = 1;
         goto finish;
     }
 
-    gPendedKextloadOperations = CFArrayCreateMutable(kCFAllocatorDefault, 0,
-        &kCFTypeArrayCallBacks);
-    if (!gPendedKextloadOperations) {
-        fprintf(stderr, "%s: memory allocation failure\n", progname);
-        exit_status = 1;
-        goto finish;
-    }
-
-    gScheduledNonsecureKexts = CFArrayCreateMutable(kCFAllocatorDefault, 0,
-        &kCFTypeArrayCallBacks);
-    if (!gScheduledNonsecureKexts) {
+    gNotifiedNonsecureKextPaths = CFDictionaryCreateMutable(kCFAllocatorDefault, 0,
+        &kCFTypeDictionaryKeyCallBacks,
+        &kCFTypeDictionaryValueCallBacks);
+    if (!gNotifiedNonsecureKextPaths) {
         fprintf(stderr, "%s: memory allocation failure\n", progname);
         exit_status = 1;
         goto finish;
@@ -291,6 +303,8 @@ int main (int argc, const char * argv[]) {
             // FIXME: ... program to trigger KLD unload?
             // FIXME: should kextd exit in this case?
         }
+
+	have_rom_mkexts = kextd_find_rom_mkexts();
     } 
 
    /*****
@@ -298,14 +312,14 @@ int main (int argc, const char * argv[]) {
     */
     CFArrayInsertValueAtIndex(repositoryDirectories, 0,
         kKXSystemExtensionsFolder);
-
    /*****
-    * Send those NDRVs down to the kernel.
+    * Make sure we scan the ROM Extensions folder.
     */
-    if (!kextd_process_ndrvs(repositoryDirectories)) {
-        // kextd_process_ndrvs() logged an error message
-        exit_status = 1;
-        goto finish;
+    if (have_rom_mkexts)
+    {
+	rom_repository_idx = 1;
+	CFArrayInsertValueAtIndex(repositoryDirectories, rom_repository_idx,
+	    CFSTR(kKXROMExtensionsFolder));
     }
 
    /*****
@@ -315,6 +329,11 @@ int main (int argc, const char * argv[]) {
         goto finish;
     }
 
+   /*****
+    * Check Extensions.mkext needs a rebuild
+    */
+    check_extensions_mkext();
+
    /*****
     * Set up the kext manager.
     */
@@ -341,6 +360,7 @@ int main (int argc, const char * argv[]) {
     KXKextManagerSetLogLevel(gKextManager, g_verbose_level);
     KXKextManagerSetLogFunction(gKextManager, kextd_log);
     KXKextManagerSetErrorLogFunction(gKextManager, kextd_error_log);
+    KXKextManagerSetWillUpdateCatalog(gKextManager, true);
 
    /*****
     * Disable clearing of relationships until we're done putting everything
@@ -369,17 +389,19 @@ int main (int argc, const char * argv[]) {
 
         result = KXKextManagerAddRepositoryDirectory(gKextManager,
             directoryURL, true /* scanForAdditions */,
-            use_repository_caches, NULL);
+            (use_repository_caches && (i != rom_repository_idx)),
+	    NULL);
         if (result != kKXKextManagerErrorNone) {
             kextd_error_log("can't add repository (%s).",
                 KXKextManagerErrorStaticCStringForError(result));
-            exit_status = 1;
-            goto finish;
         }
         CFRelease(directoryURL);
         directoryURL = NULL;
     }
 
+    CFRelease(repositoryDirectories);
+    repositoryDirectories = NULL;
+
     KXKextManagerEnableClearRelationships(gKextManager);
 
     // Create CFRunLoop & sources
@@ -415,9 +437,8 @@ finish:
     if (gMainRunLoop)                 CFRelease(gMainRunLoop);
 
 #ifndef NO_CFUserNotification
-    if (gPendedNonsecureKexts)         CFRelease(gPendedNonsecureKexts);
-    if (gPendedKextloadOperations)     CFRelease(gPendedKextloadOperations);
-    if (gScheduledNonsecureKexts)      CFRelease(gScheduledNonsecureKexts);
+    if (gPendedNonsecureKextPaths)     CFRelease(gPendedNonsecureKextPaths);
+    if (gNotifiedNonsecureKextPaths)   CFRelease(gNotifiedNonsecureKextPaths);
 #endif /* NO_CFUserNotification */
 
     exit(exit_status);
@@ -428,31 +449,183 @@ finish:
 *
 *******************************************************************************/
 
+#define TEMP_FILE		"/tmp/com.apple.iokit.kextd.XX"
+#define MKEXTUNPACK_COMMAND	"/usr/sbin/mkextunpack "	\
+				"-d "kKXROMExtensionsFolder" "
+
+static kern_return_t process_mkext(const UInt8 * bytes, CFIndex length)
+{
+    kern_return_t	err;
+    char		temp_file[1 + strlen(TEMP_FILE)];
+    char		mkextunpack_cmd[1 + strlen(TEMP_FILE) + strlen(MKEXTUNPACK_COMMAND)];
+    const char *	rom_ext_dir = kKXROMExtensionsFolder;
+    int 		outfd = -1;
+    struct stat		stat_buf;
+    
+    strcpy(temp_file, TEMP_FILE);
+    mktemp(temp_file);
+    outfd = open(temp_file, O_WRONLY|O_CREAT|O_TRUNC, 0666);
+    if (-1 == outfd) {
+	kextd_error_log("can't create %s - %s\n", temp_file,
+	    strerror(errno));
+	err = kKXKextManagerErrorFileAccess;
+	goto finish;
+    }
+
+    if (length != write(outfd, bytes, length))
+        err = kKXKextManagerErrorDiskFull;
+    else
+        err = kKXKextManagerErrorNone;
+
+    if (kKXKextManagerErrorNone != err) {
+        kextd_error_log("couldn't write output");
+	goto finish;
+    }
+
+    close(outfd);
+    outfd = -1;
+
+    if (-1 == stat(rom_ext_dir, &stat_buf))
+    {
+	if (0 != mkdir(rom_ext_dir, 0755))
+	{
+	    kextd_error_log("mkdir(%s) failed: %s\n", rom_ext_dir, strerror(errno));
+	    err = kKXKextManagerErrorFileAccess;
+	    goto finish;
+	}
+    }
+
+    strcpy(mkextunpack_cmd, MKEXTUNPACK_COMMAND);
+    strcat(mkextunpack_cmd, temp_file);
+
+    // kextd_error_log(mkextunpack_cmd);
+
+    if (0 != system(mkextunpack_cmd))
+    {
+	kextd_error_log(mkextunpack_cmd);
+	kextd_error_log("failed");
+	err = kKXKextManagerErrorChildTask;
+        goto finish;
+    }
+
+finish:
+    if (-1 != outfd)
+	close(outfd);
+    unlink(temp_file);
+
+    return err;
+}
+
+/*******************************************************************************
+*
+*******************************************************************************/
+
+static Boolean kextd_find_rom_mkexts(void)
+{
+    kern_return_t	kr;
+    CFSetRef		set = NULL;
+    CFDataRef *		mkexts = NULL;
+    CFIndex		count, idx;
+    char *		propertiesBuffer;
+    int			loaded_bytecount;
+    enum {		_kIOCatalogGetROMMkextList = 4  };
+
+    kr = IOCatalogueGetData(MACH_PORT_NULL, _kIOCatalogGetROMMkextList,
+			    &propertiesBuffer, &loaded_bytecount);
+    if (kIOReturnSuccess == kr)
+    { 
+	set = (CFSetRef)
+		IOCFUnserialize(propertiesBuffer, kCFAllocatorDefault, 0, 0);
+	vm_deallocate(mach_task_self(), (vm_address_t) propertiesBuffer, loaded_bytecount);
+    }
+    if (!set)
+	return false;
+
+    count  = CFSetGetCount(set);
+    if (count)
+    {
+	mkexts = (CFDataRef *) calloc(count, sizeof(CFDataRef));
+	CFSetGetValues(set, (const void **) mkexts);
+	for (idx = 0; idx < count; idx++)
+	{
+	    process_mkext(CFDataGetBytePtr(mkexts[idx]), CFDataGetLength(mkexts[idx]));
+	}
+	free(mkexts);
+    }
+    CFRelease(set);
+
+    return (count > 0);
+}
+
+/*******************************************************************************
+*
+*******************************************************************************/
+
+#define REBUILDMKEXT_COMMAND	"/usr/sbin/kextcache -elF -a ppc -a i386"
+
+static void check_extensions_mkext(void)
+{
+    struct stat extensions_stat_buf;
+    struct stat mkext_stat_buf;
+    Boolean rebuild;
+
+    rebuild = (0 != stat(kKXCSystemExtensionsFolder ".mkext", &mkext_stat_buf));
+    if (!rebuild && (0 == stat(kKXCSystemExtensionsFolder, &extensions_stat_buf)))
+	rebuild = (mkext_stat_buf.st_mtime != (extensions_stat_buf.st_mtime + 1));
+    
+    if (rebuild) do
+    {
+	const NXArchInfo *
+	arch = NXGetLocalArchInfo();
+	if (arch)
+	    arch = NXGetArchInfoFromCpuType(arch->cputype, CPU_SUBTYPE_MULTIPLE);
+	if (!arch)
+	{
+	    kextd_error_log("unknown architecture");
+	    break;
+	}
+
+	// kextd_error_log(REBUILDMKEXT_COMMAND);
+
+	if (0 != system(REBUILDMKEXT_COMMAND))
+	{
+	    kextd_error_log(REBUILDMKEXT_COMMAND);
+	    kextd_error_log("failed");
+	    break;
+	}
+    }
+    while (false);
+}
+
+/*******************************************************************************
+*
+*******************************************************************************/
+
 static Boolean kextd_is_running(mach_port_t * bootstrap_port_ref)
 {
     boolean_t active = FALSE;
     Boolean result = false;
     kern_return_t kern_result = KERN_SUCCESS;
-    mach_port_t   bootstrap_port;
+    mach_port_t   the_bootstrap_port;
 
     if (bootstrap_port_ref && (*bootstrap_port_ref != PORT_NULL)) {
-        bootstrap_port = *bootstrap_port_ref;
+        the_bootstrap_port = *bootstrap_port_ref;
     } else {
         /* Get the bootstrap server port */
         kern_result = task_get_bootstrap_port(mach_task_self(),
-            &bootstrap_port);
+            &the_bootstrap_port);
         if (kern_result != KERN_SUCCESS) {
             kextd_error_log("task_get_bootstrap_port(): %s\n",
                 mach_error_string(kern_result));
             exit (EX_UNAVAILABLE);
         }
         if (bootstrap_port_ref) {
-            *bootstrap_port_ref = bootstrap_port;
+            *bootstrap_port_ref = the_bootstrap_port;
         }
     }
 
     /* Check "kextd" server status */
-    kern_result = bootstrap_status(bootstrap_port,
+    kern_result = bootstrap_status(the_bootstrap_port,
         (char *)KEXTD_SERVER_NAME, &active);
     switch (kern_result) {
       case BOOTSTRAP_SUCCESS:
@@ -500,46 +673,87 @@ static int kextd_get_mach_ports(void)
 *******************************************************************************/
 int kextd_fork(void)
 {
+    sigset_t signal_set;
+    sigset_t old_signal_set;
     uid_t pid;
 
-    // prep parent to receive sigterm from child
-    signal(SIGTERM, kextd_handle_sigterm);
+    sigemptyset(&signal_set);
+    sigaddset(&signal_set, SIGTERM);
+    sigaddset(&signal_set, SIGCHLD);
+    sigaddset(&signal_set, SIGALRM);
+
+    if (sigprocmask(SIG_BLOCK, &signal_set, &old_signal_set)) {
+         kextd_error_log("sigprocmask() failed");
+         return 0;
+    }
+
+    // prep parent to receive signals from/about child
+    signal(SIGTERM, kextd_parent_handle_signal);
+    signal(SIGCHLD, kextd_parent_handle_signal);
+    signal(SIGALRM, kextd_parent_handle_signal);
 
     pid = fork();
     switch (pid) {
+
       case -1:
-        return 0;
+        return 0;  // error
         break;
+
       case 0:   // child task
+
+        // child doesn't process sigterm or sigalrm
+        signal(SIGTERM, SIG_DFL);
+        signal(SIGALRM, SIG_DFL);
+
+        if (sigprocmask(SIG_UNBLOCK, &old_signal_set, NULL)) {
+             kextd_error_log("sigprocmask() failed");
+             return 0;
+        }
+
         // Reregister/get Mach ports for the child
         if (!kextd_get_mach_ports()) {
             // kextd_get_mach_ports() logged an error message
             exit(1);
         }
 
-        // child doesn't process sigterm
-        signal(SIGTERM, SIG_DFL);
         // FIXME: old kextd did this; is it needed?
         _CFRunLoopSetCurrent(NULL);
         break;
-      default:  // parent task
-        {
-            /* parent: wait for signal, then exit */
-            int status;
 
-            wait4(pid, (int *)&status, 0, 0);
-            if (WIFEXITED(status)) {
-                kextd_error_log(
-                    "*** %s (daemon) failed to start, exit status=%d",
-                    progname, WEXITSTATUS(status));
-            } else {
-                kextd_error_log(
-                    "*** %s (daemon) failed to start, received signal=%d",
-                    progname, WTERMSIG(status));
-            }
-            fflush (stderr);
-            exit(1);
-        }
+      default:  // parent task
+	    kextd_openlog("kextd-parent");  // should that arg be progname?
+
+		struct itimerval timer_buf;
+		timer_buf.it_interval.tv_sec = 0;
+		timer_buf.it_interval.tv_usec = 0;
+		timer_buf.it_value.tv_sec = 40;
+		timer_buf.it_value.tv_usec = 0;
+		setitimer(ITIMER_REAL, &timer_buf, NULL);
+
+		while (1) {
+			sigsuspend(&old_signal_set);
+			if (parent_received_sigterm) {
+				kern_return_t    kern_result;
+				mach_timespec_t  waitTime = { 40, 0 };
+
+				kern_result = IOKitWaitQuiet(g_io_master_port, &waitTime);
+				if (kern_result == kIOReturnTimeout) {
+					kextd_error_log("IOKitWaitQuiet() timed out",
+					kern_result);
+				} else if (kern_result != kIOReturnSuccess) {
+					kextd_error_log("IOKitWaitQuiet() failed with result code %lx",
+					kern_result);
+					exit(1);
+				}
+				exit(0);
+			} else if (parent_received_sigchld) {
+				kextd_error_log("kextd forked child task exited abnormally");
+				exit(1);
+			} else if (parent_received_sigalrm) {
+				kextd_error_log("kextd parent task timed out waiting for signal from child");
+				exit(1);
+			}
+		}
         break;
     }
 
@@ -563,6 +777,7 @@ int kextd_fork(void)
 /*******************************************************************************
 * kextd_set_up_server()
 *******************************************************************************/
+// in mig_server.c
 extern void kextd_mach_port_callback(
     CFMachPortRef port,
     void *msg,
@@ -576,8 +791,17 @@ static Boolean kextd_set_up_server(void)
     CFRunLoopSourceContext sourceContext;
     unsigned int sourcePriority = 1;
     CFMachPortRef kextdMachPort = NULL;  // must release
+    mach_port_limits_t limits;  // queue limit for signal-handler port
 
-    if (kextd_is_running(&bootstrap_port)) {
+    if (kextd_is_running(&gBootstrap_port)) {
+        result = false;
+        goto finish;
+    }
+
+    gRunLoopSourceLock = PTLockCreate();
+    if (!gRunLoopSourceLock) {
+        kextd_error_log(
+            "failed to create kernel request run loop source lock");
         result = false;
         goto finish;
     }
@@ -598,7 +822,7 @@ static Boolean kextd_set_up_server(void)
     * It's important that each source have a distinct priority; sharing
     * them causes unpredictable behavior with the runloop.
     */
-    sourceContext.perform = kextd_handle_sighup_in_runloop;
+    sourceContext.perform = kextd_handle_rescan;
     gRescanRunLoopSource = CFRunLoopSourceCreate(kCFAllocatorDefault,
         sourcePriority++, &sourceContext);
     if (!gRescanRunLoopSource) {
@@ -632,17 +856,38 @@ static Boolean kextd_set_up_server(void)
     CFRunLoopAddSource(gMainRunLoop, gClientRequestRunLoopSource,
         kCFRunLoopDefaultMode);
 
-#ifndef NO_CFUserNotification
-
-    sourceContext.perform = kextd_handle_pended_kextload;
-    gNonsecureKextRunLoopSource = CFRunLoopSourceCreate(kCFAllocatorDefault,
-        sourcePriority++, &sourceContext);
-    if (!gNonsecureKextRunLoopSource) {
-       kextd_error_log("couldn't create pended kextload run loop source");
+    gKextdSignalMachPort = CFMachPortCreate(kCFAllocatorDefault,
+        kextd_runloop_handle_sighup, NULL, NULL);
+    limits.mpl_qlimit = 1;
+    kern_result = mach_port_set_attributes(mach_task_self(),
+        CFMachPortGetPort(gKextdSignalMachPort),
+        MACH_PORT_LIMITS_INFO,
+        (mach_port_info_t)&limits,
+        MACH_PORT_LIMITS_INFO_COUNT);
+    if (kern_result != KERN_SUCCESS) {
+        kextd_error_log("failed to set signal-handling port limits");
+    }
+    gSignalRunLoopSource = CFMachPortCreateRunLoopSource(
+        kCFAllocatorDefault, gKextdSignalMachPort, sourcePriority++);
+    if (!gSignalRunLoopSource) {
+       kextd_error_log("couldn't create signal-handling run loop source");
         result = false;
         goto finish;
     }
-    CFRunLoopAddSource(gMainRunLoop, gNonsecureKextRunLoopSource,
+    CFRunLoopAddSource(gMainRunLoop, gSignalRunLoopSource,
+        kCFRunLoopDefaultMode);
+
+#ifndef NO_CFUserNotification
+
+    sourceContext.perform = kextd_check_notification_queue;
+    gNotificationQueueRunLoopSource = CFRunLoopSourceCreate(kCFAllocatorDefault,
+        sourcePriority++, &sourceContext);
+    if (!gNotificationQueueRunLoopSource) {
+       kextd_error_log("couldn't create alert run loop source");
+        result = false;
+        goto finish;
+    }
+    CFRunLoopAddSource(gMainRunLoop, gNotificationQueueRunLoopSource,
         kCFRunLoopDefaultMode);
 
 #endif /* NO_CFUserNotification */
@@ -676,7 +921,7 @@ finish:
     if (gKernelRequestRunLoopSource)  CFRelease(gKernelRequestRunLoopSource);
     if (gClientRequestRunLoopSource)  CFRelease(gClientRequestRunLoopSource);
 #ifndef NO_CFUserNotification
-    if (gNonsecureKextRunLoopSource)  CFRelease(gNonsecureKextRunLoopSource);
+    if (gNotificationQueueRunLoopSource) CFRelease(gNotificationQueueRunLoopSource);
 #endif /* NO_CFUserNotification */
     if (kextdMachPort)                CFRelease(kextdMachPort);
 
@@ -706,27 +951,64 @@ void kextd_register_signals(void)
 * registered and used by parent of forked daemon to exit
 * upon signal from forked daemon.
 *******************************************************************************/
-void kextd_handle_sigterm(int signum)
+void kextd_parent_handle_signal(int signum)
 {
-    kern_return_t    kern_result;
-    mach_timespec_t  waitTime = { 40, 0 };
-
-    kern_result = IOKitWaitQuiet(g_io_master_port, &waitTime);
-    if (kern_result == kIOReturnTimeout) {
-        kextd_error_log("IOKitWaitQuiet() timed out",
-        kern_result);
-    } else if (kern_result != kIOReturnSuccess) {
-        kextd_error_log("IOKitWaitQuiet() failed with result code %lx",
-        kern_result);
+    switch (signum) {
+      case SIGTERM:
+        parent_received_sigterm = true;
+        break;
+      case SIGCHLD:
+        parent_received_sigchld = true;
+        break;
+      case SIGALRM:
+        parent_received_sigalrm = true;
+        break;
     }
-    exit(0);
+    return;
+}
+
+/*******************************************************************************
+* On receiving a SIGHUP, the daemon sends a Mach message to the signal port,
+* causing the run loop handler function kextd_runloop_handle_sighup() to be called.
+*******************************************************************************/
+void kextd_handle_sighup(int signum)
+{
+    mach_msg_empty_send_t msg;
+    mach_msg_option_t options;
+    kern_return_t kern_result;
+
+    if (signum != SIGHUP) {
+        return;
+    }
+
+   /*
+    * send message to indicate that a request has been made
+    * for the daemon to be shutdown.
+    */
+    msg.header.msgh_bits = MACH_MSGH_BITS(MACH_MSG_TYPE_COPY_SEND, 0);
+    msg.header.msgh_size = sizeof(msg);
+    msg.header.msgh_remote_port = CFMachPortGetPort(gKextdSignalMachPort);
+    msg.header.msgh_local_port = MACH_PORT_NULL;
+    msg.header.msgh_id = 0;
+    options = MACH_SEND_TIMEOUT;
+    kern_result = mach_msg(&msg.header,  /* msg */
+        MACH_SEND_MSG|options,      /* options */
+        msg.header.msgh_size,       /* send_size */
+        0,                          /* rcv_size */
+        MACH_PORT_NULL,             /* rcv_name */
+        0,                          /* timeout */
+        MACH_PORT_NULL);            /* notify */
     return;
 }
 
 /*******************************************************************************
 *
 *******************************************************************************/
-void kextd_handle_sighup(int signum)
+void kextd_runloop_handle_sighup(
+    CFMachPortRef port,
+    void *msg,
+    CFIndex size,
+    void *info)
 {
     if (gRescanRunLoopSource) {
         PTLockTakeLock(gRunLoopSourceLock);
@@ -737,40 +1019,49 @@ void kextd_handle_sighup(int signum)
     } else {
         kextd_log("received SIGHUP before entering run loop; ignoring");
     }
+    return;
+}
 
+#ifndef NO_CFUserNotification
+/*******************************************************************************
+*
+*******************************************************************************/
+void kextd_clear_all_notifications(void)
+{
+    CFArrayRemoveAllValues(gPendedNonsecureKextPaths);
+
+   /* Release any reference to the current user notification.
+    */
+    if (gCurrentNotification) {
+        CFUserNotificationCancel(gCurrentNotification);
+        CFRelease(gCurrentNotification);
+        gCurrentNotification = NULL;
+    }
+
+    if (gCurrentNotificationRunLoopSource) {
+        CFRunLoopRemoveSource(gMainRunLoop, gCurrentNotificationRunLoopSource,
+            kCFRunLoopDefaultMode);
+        CFRelease(gCurrentNotificationRunLoopSource);
+        gCurrentNotificationRunLoopSource = NULL;
+    }
+
+   /* Clear the record of which kexts the user has been told are insecure.
+    * They'll get all the same warnings upon logging in again.
+    */
+    CFDictionaryRemoveAllValues(gNotifiedNonsecureKextPaths);
 
     return;
 }
+#endif /* NO_CFUserNotification */
+
 
 /*******************************************************************************
 *
 *******************************************************************************/
-void kextd_rescan(void)
+void kextd_handle_rescan(void * info)
 {
 #ifndef NO_CFUserNotification
-   /* Dump all nonsecure kexts awaiting load by kextd. We're rescanning so
-    * this will all get retriggered.
-    */
-    CFArrayRemoveAllValues(gPendedNonsecureKexts);
-    CFArrayRemoveAllValues(gScheduledNonsecureKexts);
-
-   /* Clear any security alert awaiting user input. Don't clear a kextload
-    * operation awaiting user input, however, as that would likely never
-    * get retriggered, and it doesn't access the kextmanager data set.
-    */
-    if (gSecurityNotification) {
-        CFUserNotificationCancel(gSecurityNotification);
-        CFRelease(gSecurityNotification);
-        gSecurityNotification = NULL;
-    }
-    if (gFailureNotification) {
-        CFUserNotificationCancel(gFailureNotification);
-        CFRelease(gFailureNotification);
-        gFailureNotification = NULL;
-    }
-    gSecurityAlertKext = NULL;
-    gResendSecurityAlertKextPersonalities = false;
-
+    kextd_clear_all_notifications();
 #endif /* NO_CFUserNotification */
 
     KXKextManagerResetAllRepositories(gKextManager);
@@ -781,51 +1072,16 @@ void kextd_rescan(void)
     return;
 }
 
-/*******************************************************************************
-*
-*******************************************************************************/
-void kextd_handle_sighup_in_runloop(void * info)
-{
-    kextd_rescan();
-    return;
-}
-
 /*******************************************************************************
 *
 *******************************************************************************/
 static Boolean kextd_download_personalities(void)
 {
-    Boolean result = true;
-    CFArrayRef allKextPersonalities = NULL;  // must release
+    KXKextManagerError result;
 
-   /*****
-    * Empty the kernel's catalogue and send all candidate kext personalities
-    * down.
-    */
-    IOCatalogueReset(g_io_master_port, kIOCatalogResetDefault);
+    result = KXKextManagerSendAllKextPersonalitiesToCatalog(gKextManager);
 
-    allKextPersonalities = KXKextManagerCopyAllKextPersonalities(gKextManager);
-    if (!allKextPersonalities) {
-        kextd_error_log("can't get kext personalities to send to kernel");
-        result = false;
-        goto finish;
-    }
-
-    if (KXKextManagerSendPersonalitiesToCatalog(gKextManager,
-           allKextPersonalities) != kKXKextManagerErrorNone) {
-
-        kextd_error_log("can't send kext personalities to kernel");
-        result = false;
-        goto finish;
-    }
-
-finish:
-
-    if (allKextPersonalities) {
-        CFRelease(allKextPersonalities);
-    }
-
-    return result;
+    return (kKXKextManagerErrorNone == result) ? true : false;
 }
 
 /*******************************************************************************
@@ -870,7 +1126,7 @@ static Boolean kextd_process_ndrvs(CFArrayRef repositoryDirectories)
             goto finish;
         }
 
-	IOLoadPEFsFromURL( ndrvDirURL, MACH_PORT_NULL );
+        IOLoadPEFsFromURL( ndrvDirURL, MACH_PORT_NULL );
     }
 
 finish:
