@@ -19,12 +19,9 @@ Boolean kextd_launch_kernel_request_thread(void);
 void * kextd_kernel_request_loop(void * arg);
 
 void kextd_handle_kernel_request(void * info);
+void kextd_check_notification_queue(void * info);
 
 void kextd_load_kext(char * kmod_name,
     KXKextManagerError * result /* out */);
 
-#ifndef NO_CFUserNotification
-void kextd_handle_pended_kextload(void * info);
-#endif /* NO_CFUserNotification */
-
 #endif __REQUEST_H__
